/* $Id: stubs.c,v 1.2 1999/01/12 20:06:30 becka Exp $
******************************************************************************

   Bse extension stubs.

   Copyright (C) 1999 Andreas Beck - becka@ggi-project.org
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>

#include "bse_int.h"

/* Allocating/Destroying a BSE.
 */
ggibse_t	ggiBseAllocate	(ggi_visual_t vis, int width, int height, enum bsetype type)
{
	ggibse_t	newsprite;
	ggibse_t	current;

	if (NULL==(newsprite=malloc(sizeof(*newsprite)))) 
		return NULL;

	newsprite->vis=vis;
	newsprite->x  =
	newsprite->y  =0;
	newsprite->w  =width;
	newsprite->h  =height;
	if (type==BSE_BOB)
		newsprite->hidecount=0;	/* BOBs have no notion of being hidden */
	else	newsprite->hidecount=1;	/* come up hidden ! */
	newsprite->type=type;
	newsprite->image=
	newsprite->alpha=
	newsprite->backgr=NULL;

	if (LIBGGI_BSEEXT(newsprite->vis)->alloc(newsprite)) {
		free(newsprite);
		return NULL;
	}

	current=LIBGGI_BSEEXT(newsprite->vis)->first;

	while(current&&current->next) current=current->next;

	newsprite->prev=current;
	if (current) current->next=newsprite; 
	else	LIBGGI_BSEEXT(newsprite->vis)->first=newsprite;
	newsprite->next=NULL;

	return newsprite;
}

int		ggiBseFree	(ggibse_t bob)
{
	if (!bob) return 0;	/* Sprite wasn't alloced. Keep it simple. */

	/* Unlink it. */
	if (bob->prev) bob->prev->next=bob->next;
	else LIBGGI_BSEEXT(bob->vis)->first=bob->next;
	if (bob->next) bob->next->prev=bob->prev;

	if (bob->image) free(bob->image);
	if (bob->alpha) free(bob->alpha);
	if (bob->backgr) free(bob->backgr);
	free(bob);
	return 0;
}

/* Hiding and showing
 */
void ggiBseHideAboveCollide(ggibse_t bob,int x,int y) 
{
	ggibse_t curr=bob;
	while((curr=curr->prev)) {
		if (curr->type==BSE_EMUSPRITE) {
			if (curr->x+curr->w<=x ||
			    curr->y+curr->h<=y ||
			          x+bob->w <=curr->x ||
			          y+bob->h <=curr->y ) continue; 
			ggiBseHideAboveCollide(curr,curr->x,curr->y);
			ggiBseHide(curr);
		}
	}
}

void ggiBseShowAboveCollide(ggibse_t bob,int x,int y) 
{
	ggibse_t curr=bob;
	while((curr=curr->prev)) {
		if (curr->type==BSE_EMUSPRITE) {
			if (curr->x+curr->w<=x ||
			    curr->y+curr->h<=y ||
			          x+bob->w <=curr->x ||
			          y+bob->h <=curr->y ) continue; 
			ggiBseShow(curr);
			ggiBseShowAboveCollide(curr,curr->x,curr->y);
		}
	}
}



/* Populate the sprite
 */
int ggiBseSetImage	(ggibse_t bob,ggi_visual_t vis,int x,int y,
			 unsigned char *alpha)
{
	return bob->setimage(bob,vis,x,y,alpha);
}

/* Move the sprite around.
 */
int ggiBseMove		(ggibse_t bob,int x,int y)
{
	int rc,oldx,oldy;
	
	if (bob->hidecount) {
		bob->x=x;
		bob->y=y;
		return 0;	/* It's hidden. */
	}
	oldx=bob->x;oldy=bob->y;

	if (bob->type==BSE_EMUSPRITE) {
		ggiBseHideAboveCollide(bob,oldx,oldy);
		ggiBseHideAboveCollide(bob,x,y);
	}

	rc=bob->move(bob,x,y);

	if (bob->type==BSE_EMUSPRITE) {
		ggiBseShowAboveCollide(bob,x,y);
		ggiBseShowAboveCollide(bob,oldx,oldy);
	}
	return rc;
}

int ggiBseHide		(ggibse_t bob)
{
	int rc;

	if (bob->type==BSE_BOB)		/* A BOB cannot be hidden. */
		return GGI_EARGINVAL;
	if (bob->hidecount++) return 0;	/* It's already hidden. */

	if (bob->type==BSE_EMUSPRITE)
		ggiBseHideAboveCollide(bob,bob->x,bob->y);

	rc=bob->show(bob,!bob->hidecount);	/* Hide it. */

	if (bob->type==BSE_EMUSPRITE)
		ggiBseShowAboveCollide(bob,bob->x,bob->y);

	return rc;
}

int ggiBseShow		(ggibse_t bob)
{
	int rc;
	
	if (bob->type==BSE_BOB)		/* A BOB cannot be hidden. */
		return GGI_EARGINVAL;

	if (bob->hidecount<=0) return GGI_EARGINVAL;	/* Too often ! */

	if (--(bob->hidecount)==0) {

		if (bob->type==BSE_EMUSPRITE)
			ggiBseHideAboveCollide(bob,bob->x,bob->y);

		rc=bob->show(bob,!bob->hidecount);	/* Make visible */

		if (bob->type==BSE_EMUSPRITE)
			ggiBseShowAboveCollide(bob,bob->x,bob->y);

		return rc;
	}
	return 0;	/* We are still hidden. */
}

/* These are for EMU_SPRITEs. You can use these for hiding them for drawing.
 */
int ggiBseSoftHide	(ggibse_t bob)
{
	switch(bob->type) {
		case BSE_BOB:
		default:
			return GGI_EARGINVAL;
		case BSE_SPRITE:
			return 0;	/* NOP for real sprites */
		case BSE_EMUSPRITE:
			return ggiBseHide(bob);
	}
}

int ggiBseSoftShow	(ggibse_t bob)
{
	switch(bob->type) {
		case BSE_BOB:
		default:
			return GGI_EARGINVAL;
		case BSE_SPRITE:
			return 0;	/* NOP for real sprites */
		case BSE_EMUSPRITE:
			return ggiBseShow(bob);
	}
}

/* Helper functions for Alpha channels
 */
 
unsigned char *ggiBseMakeAlphaFromPixelkey(ggibse_t bob,
	ggi_visual_t vis,int x,int y,ggi_pixel pix) {

	unsigned char *rc,*alph;
	ggi_pixel gp;
	int xx,yy;

	if ((rc=malloc(bob->w*bob->h*sizeof(unsigned char)))) {
		alph=rc;
		for(yy=0;yy<bob->h;yy++)
			for(xx=0;xx<bob->w;xx++) {
				ggiGetPixel(vis,x+xx,y+yy,&gp);
				*alph++=(gp==pix) ? 0xff : 0x00;
			}
	}
	return rc;
}

unsigned char *ggiBseMakeAlphaFromColorkey(ggibse_t bob,
	ggi_visual_t vis,int x,int y,ggi_color *col) {
	return ggiBseMakeAlphaFromPixelkey(bob,vis,x,y,ggiMapColor(vis,col));
}

static unsigned char func_treshold(unsigned char inval,void *tresh)
{
	return inval > *((unsigned char *)tresh) ? 0xff : 0x00;
}

static unsigned char func_invert(unsigned char inval,void *unused)
{
	return 0xff-inval;
}

unsigned char *ggiBseChangeAlphaByFunction(ggibse_t bob, 
	unsigned char *alpha,
	unsigned char (*func)(unsigned char in,void *parms),void *parms) {

	int xx,yy;

	for(yy=0;yy<bob->h;yy++)
		for(xx=0;xx<bob->w;xx++) {
			*alpha=func(*alpha,parms);
		}

	return alpha;
}

unsigned char *ggiBseChangeAlphaTreshold(ggibse_t bob, 
	unsigned char *alpha, unsigned char treshold) {
	return ggiBseChangeAlphaByFunction(bob,alpha,func_treshold,&treshold);
}

unsigned char *ggiBseChangeAlphaInvert(ggibse_t bob, 
	unsigned char *alpha) {
	return ggiBseChangeAlphaByFunction(bob,alpha,func_invert,NULL);
}
